/*
 * This code contains functions that use the "recursive backtracking"
 * algorithmic strategy.  There are functions to find all the permutations
 * and combinations of possible orderings of the letters of a string.
 * There are also functions for rolling dice and finding all combinations of
 * dice values or all combinations that add up to a particular sum.
 */

#include <fstream>
#include <iomanip>
#include <iostream>
#include <set>
#include <stack>

using namespace std;

// function prototype declarations
void permute(string s, string chosen = "");
void combinations(string s, int length);
void combinationsHelper(string s, int length, string chosen, set<string>& alreadyPrinted);
void diceRolls(int dice);
void diceRollsHelper(int dice, stack<int>& chosen);
void diceSum(int dice, int desiredSum);
void diceSumHelper(int dice, int desiredSum, int sum, stack<int>& chosen);
void printStack(stack<int> s);

int main() {
    cout << "permute(\"MARTY\"):" << endl;
    permute("MARTY");
    cout << endl;

    cout << "combinations(\"GOOGLE\", 3):" << endl;
    combinations("GOOGLE", 3);
    cout << endl;

    cout << "diceRolls(3):" << endl;
    diceRolls(3);
    cout << endl;

    cout << "diceSum(3, 7):" << endl;
    diceSum(3, 7);
    cout << endl;

    return 0;
}

/*
 * Prints all possible rearrangements of the letters of the given string.
 * The optional 'chosen' parameter contains all letters that have been chosen
 * so far by previous calls.
 * The algorithm works by choosing a letter at a time and then recurring,
 * until there is nothing left to choose.  Then it prints the arrangement
 * chosen, and the code returns ("backtracks") to a prior call so it can
 * explore further possibilities.
 */
void permute(string s, string chosen) {
    if (s == "") {
        // base case; nothing to choose, so print the word chosen
        cout << chosen << endl;
    } else {
        // recursive case; for each choice:
        for (int i = 0; i < s.length(); i++) {
            permute(s.substr(0, i) + s.substr(i + 1), chosen + s[i]);
        }
    }
}

/*
 * Prints all possible rearrangements of UNIQUE letters of the given string
 * that are exactly the given length.
 */
void combinations(string s, int length) {
    set<string> alreadyPrinted;
    combinationsHelper(s, length, "", alreadyPrinted);
}

/*
 * Recursive helper to implement combinations behavior.
 * The 'chosen' parameter contains all letters that have been chosen
 * so far by previous calls, and the 'alreadyPrinted' parameter remembers
 * all strings that have been printed before so it doesn't print the same
 * string twice.
 * The algorithm works by choosing a letter at a time and then recurring,
 * until 'length' unique letters have been chosen.  Then it prints the
 * arrangement chosen, and the code returns ("backtracks") to a prior call
 * so it can explore further possibilities.
 */
void combinationsHelper(string s, int length, string chosen, set<string> &alreadyPrinted) {
    if (length == 0) {
        if (alreadyPrinted.find(chosen) == alreadyPrinted.end()) {
            cout << chosen << endl;
            alreadyPrinted.insert(chosen);
        }
    } else {
        // recursive case; for each choice:
        for (int i = 0; i < (int) s.length(); i++) {
            // choose
            char c = s[i];
            if (chosen.find(c) == string::npos) {
                s.erase(i, 1);   // s = "ATY"
                chosen += c;     // c = "MR"

                // explore
                combinationsHelper(s, length - 1, chosen, alreadyPrinted);

                // unchoose
                s.insert(i, 1, c);  // s = "ARTY"
                chosen.erase(chosen.length() - 1, 1);
            }
        }
    }
}

/*
 * Prints all possible combinations of dice values that can be rolled using
 * the given number of 6-sided dice,  one combination per line.
 */
void diceRolls(int dice) {
    stack<int> chosen;
    diceRollsHelper(dice, chosen);
}

/*
 * Recursive helper that uses backtracking to explore the given number of
 * remaining dice to be rolled.
 */
void diceRollsHelper(int dice, stack<int>& chosen) {
    if (dice == 0) {
        // base case: all dice have been chosen, so print
        printStack(chosen);
        cout << endl;
    } else {
        // recursive case: for each possible choice, choose-explore-unchoose
        // the "choices" here are, what should the next die be? (1-6)
        for (int i = 1; i <= 6; i++) {
            chosen.push(i);                      // choose
            diceRollsHelper(dice - 1, chosen);   // explore
            chosen.pop();                        // un-choose
        }
    }
}

/*
 * Prints all possible combinations of dice values that can be rolled using
 * the given number of 6-sided dice that add up to the given desired sum,
 * one combination per line.
 */
void diceSum(int dice, int desiredSum) {
    stack<int> chosen;
    diceSumHelper(dice, desiredSum, 0, chosen);
}

/*
 * Recursive helper that uses backtracking to explore the given number of
 * remaining dice to be rolled.
 * The 'sum' parameter tracks the sum of all dice rolled so far, and
 * the 'chosen' stack holds all dice that have been chosen.
 *
 * The code contains a few optimizations to avoid making recursive calls that
 * have no chance of producing a successful outcome.  This is called 'pruning'
 * the call tree, and it reduces the number of recursive calls needed.
 */
void diceSumHelper(int dice, int desiredSum, int sum, stack<int>& chosen) {
    if (dice == 0) {
        if (sum == desiredSum) {
            // base case: found exactly the right sum, so print the dice
            printStack(chosen);
            cout << endl;
        }
    } else {
        // recursive case: what should the next die be?  1-6
        for (int i = 1; i <= 6; i++) {                  // optimization:
            if (sum + i + 1*(dice-1) <= desiredSum &&   // - not too big
                sum + i + 6*(dice-1) >= desiredSum) {   // - not too small
                // choose
                chosen.push(i);

                // explore
                diceSumHelper(dice - 1, desiredSum, sum + i, chosen);

                // un-choose
                chosen.pop();
            }
        }
    }
}

void printStack(stack<int> s) {
    while (!s.empty()) {
        cout << s.top();
        s.pop();
    }
}
